# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import operator
from datetime import date, datetime, time, timedelta
from typing import Callable, Dict, Optional

import shapely

from .. import ast, values
from ..util import like_pattern_to_re
from .evaluator import Evaluator, handle

COMPARISON_MAP = {
    "=": operator.eq,
    "<>": operator.ne,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
}

ARITHMETIC_MAP = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
}


def is_literal(value):
    return isinstance(value, values.LITERALS)


TEMPORAL_LITERALS = (date, datetime, time, timedelta, values.Interval)


def is_temporal_literal(value):
    return isinstance(value, TEMPORAL_LITERALS)


GEOMETRY_LITERALS = (values.Geometry, values.Envelope)


def is_geometry_literal(value):
    return isinstance(value, GEOMETRY_LITERALS)


def is_any_literal(value):
    return is_literal(value) or is_temporal_literal(value) or is_geometry_literal(value)


def to_geometry(value):
    if isinstance(value, values.Geometry):
        return shapely.geometry.shape(value)
    elif isinstance(value, values.Envelope):
        return shapely.geometry.Polygon.from_bounds(
            value.x1, value.y1, value.x2, value.y2
        )
    raise ValueError(str(type(value)))


class OptimizeEvaluator(Evaluator):
    def __init__(self, function_map: Dict[str, Callable]):
        self.function_map = function_map

    @handle(ast.Not)
    def not_(self, node, sub):
        if isinstance(sub, bool):
            return operator.not_(sub)
        else:
            return ast.Not(sub)

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        if isinstance(lhs, bool) and isinstance(rhs, bool):
            op = operator.and_ if node.op.value == "AND" else operator.or_
            return op(lhs, rhs)

        elif isinstance(lhs, bool) or isinstance(rhs, bool):
            if isinstance(lhs, bool):
                certain, uncertain = lhs, rhs
            else:
                certain, uncertain = rhs, lhs

            # for OR nodes, when we have one true branch, the other
            # can be dropped. Otherwise we can shorthand to the
            # uncertain branch
            if node.op.value == "OR":
                if certain:
                    return True
                else:
                    return uncertain
            # for AND nodes, we can drop the node if the certain one is
            # false. Otherwise we can shorthand to the other
            elif node.op.value == "AND":
                if certain:
                    return uncertain
                else:
                    return False
            # we can eliminate the whole node and its sub-nodes, as it
            # will always evaluate to false
            return False

        else:
            return (ast.And if node.op.value == "AND" else ast.Or)(lhs, rhs)

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        if is_literal(lhs) and is_literal(rhs):
            op = COMPARISON_MAP[node.op.value]
            return op(lhs, rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        if is_literal(lhs) and is_literal(low) and is_literal(high):
            result = low <= lhs <= high
            if node.not_:
                result = not result
            return result
        else:
            return ast.Between(lhs, low, high, node.not_)

    @handle(ast.Like)
    def like(self, node, lhs):
        if is_literal(lhs):
            regex = like_pattern_to_re(
                node.pattern,
                node.nocase,
                node.wildcard,
                node.singlechar,
                node.escapechar,
            )
            result = regex.match(lhs) is not None
            if node.not_:
                result = not result
            return result
        else:
            return ast.Like(
                lhs,
                node.pattern,
                node.nocase,
                node.wildcard,
                node.singlechar,
                node.escapechar,
                node.not_,
            )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        if is_literal(lhs) and all(is_literal(o) for o in options):
            result = lhs in options
            if node.not_:
                result = not result
            return result
        else:
            return ast.In(lhs, list(options), node.not_)

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return ast.IsNull(lhs, node.not_)

    @handle(ast.Exists)
    def exists(self, node, lhs):
        return ast.Exists(lhs, node.not_)

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        if is_temporal_literal(lhs) and is_temporal_literal(rhs):
            lhs = to_interval(lhs)
            rhs = to_interval(rhs)

            return node.op.value == relate_intervals(lhs, rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.ArrayPredicate, subclasses=True)
    def array(self, node, lhs, rhs):
        if isinstance(lhs, list) and isinstance(rhs, list):
            left = set(lhs)
            right = set(rhs)

            if node.op == ast.ArrayComparisonOp.AEQUALS:
                return left == right
            elif node.op == ast.ArrayComparisonOp.ACONTAINS:
                return left >= right
            elif node.op == ast.ArrayComparisonOp.ACONTAINEDBY:
                return left <= right
            elif node.op == ast.ArrayComparisonOp.AOVERLAPS:
                return bool(left & right)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        if is_geometry_literal(lhs) and is_geometry_literal(rhs):
            lhs = to_geometry(lhs)
            rhs = to_geometry(rhs)
            op = getattr(lhs, node.op.value.lower())
            return op(rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.Relate)
    def spatial_pattern(self, node, lhs, rhs):
        if is_geometry_literal(lhs) and isinstance(rhs, str):
            lhs = to_geometry(lhs)
            return lhs.relate_pattern(rhs, node.pattern)
        else:
            return ast.Relate(lhs, rhs, node.pattern)

    @handle(ast.SpatialDistancePredicate)
    def distance(self, node, lhs, rhs):
        # TODO: can this be reduced?
        return type(node)(lhs, rhs, node.distance, node.units)

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        if is_geometry_literal(lhs):
            lhs = to_geometry(lhs)
            return lhs.intersects(
                shapely.geometry.Polygon.from_bounds(
                    node.minx, node.miny, node.maxx, node.maxy
                )
            )
        else:
            return ast.BBox(lhs, node.minx, node.miny, node.maxx, node.maxy)

    @handle(ast.Attribute)
    def attribute(self, node):
        return node

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        if is_literal(lhs) and is_literal(rhs):
            op = ARITHMETIC_MAP[node.op.value]
            return op(lhs, rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.Function)
    def function(self, node, *arguments):
        func = self.function_map.get(node.name)
        if func and all(is_any_literal(a) for a in arguments):
            return func(*arguments)
        else:
            return ast.Function(node.name, list(arguments))

    # just pass through these nodes
    @handle(ast.Attribute, values.Geometry, values.Envelope, *values.LITERALS)
    def literal(self, node):
        return node


def to_interval(value):
    # TODO:
    zulu = None
    if isinstance(value, values.Interval):
        low = value.start
        high = value.end
        if isinstance(low, date):
            low = datetime.combine(low, time.min, zulu)
        if isinstance(high, date):
            high = datetime.combine(high, time.max, zulu)

        if isinstance(low, timedelta):
            low = high - timedelta
        elif isinstance(high, timedelta):
            high = low + timedelta

        return (low, high)

    elif isinstance(value, date):
        return (
            datetime.combine(value, time.min, zulu),
            datetime.combine(value, time.max, zulu),
        )

    elif isinstance(value, datetime):
        return (value, value)

    raise ValueError(f"Invalid type {type(value)}")


def relate_intervals(lhs, rhs):  # noqa: C901
    ll, lh = lhs
    rl, rh = rhs
    if lh < rl:
        return "BEFORE"
    elif ll > rh:
        return "AFTER"
    elif lh == rl:
        return "MEETS"
    elif ll == rh:
        return "METBY"
    elif ll < rl and rl < lh < rh:
        return "TOVERLAPS"
    elif rl < ll < rh and lh > rh:
        return "OVERLAPPEDBY"
    elif ll == rl and lh < rh:
        return "BEGINS"
    elif ll == rl and lh > rh:
        return "BEGUNBY"
    elif ll > rl and lh < rh:
        return "DURING"
    elif ll < rl and lh > rh:
        return "TCONTAINS"
    elif ll > rl and lh == rh:
        return "TENDS"
    elif ll < rl and lh == rh:
        return "ENDEDBY"
    elif ll == rl and lh == rh:
        return "TEQUALS"

    raise ValueError(f"Error relating intervals [{ll}, {lh}] and ({rl}, {rh})")


def optimize(
    root: ast.Node, function_map: Optional[Dict[str, Callable]] = None
) -> ast.Node:
    result = OptimizeEvaluator(function_map or {}).evaluate(root)
    if isinstance(result, bool):
        result = ast.Include(not result)

    return result



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import Dict, Optional

import shapely.geometry

from ... import ast, values
from ..evaluator import Evaluator, handle

COMPARISON_OP_MAP = {
    ast.ComparisonOp.EQ: "=",
    ast.ComparisonOp.NE: "<>",
    ast.ComparisonOp.LT: "<",
    ast.ComparisonOp.LE: "<=",
    ast.ComparisonOp.GT: ">",
    ast.ComparisonOp.GE: ">=",
}


ARITHMETIC_OP_MAP = {
    ast.ArithmeticOp.ADD: "+",
    ast.ArithmeticOp.SUB: "-",
    ast.ArithmeticOp.MUL: "*",
    ast.ArithmeticOp.DIV: "/",
}

SPATIAL_COMPARISON_OP_MAP = {
    ast.SpatialComparisonOp.INTERSECTS: "ST_Intersects",
    ast.SpatialComparisonOp.DISJOINT: "ST_Disjoint",
    ast.SpatialComparisonOp.CONTAINS: "ST_Contains",
    ast.SpatialComparisonOp.WITHIN: "ST_Within",
    ast.SpatialComparisonOp.TOUCHES: "ST_Touches",
    ast.SpatialComparisonOp.CROSSES: "ST_Crosses",
    ast.SpatialComparisonOp.OVERLAPS: "ST_Overlaps",
    ast.SpatialComparisonOp.EQUALS: "ST_Equals",
}


class SQLEvaluator(Evaluator):
    def __init__(self, attribute_map: Dict[str, str], function_map: Dict[str, str]):
        self.attribute_map = attribute_map
        self.function_map = function_map

    @handle(ast.Not)
    def not_(self, node, sub):
        return f"NOT {sub}"

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        return f"({lhs} {node.op.value} {rhs})"

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        return f"({lhs} {COMPARISON_OP_MAP[node.op]} {rhs})"

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        return f"({lhs} {'NOT ' if node.not_ else ''}BETWEEN {low} AND {high})"

    @handle(ast.Like)
    def like(self, node, lhs):
        pattern = node.pattern
        if node.wildcard != "%":
            # TODO: not preceded by escapechar
            pattern = pattern.replace(node.wildcard, "%")
        if node.singlechar != "_":
            # TODO: not preceded by escapechar
            pattern = pattern.replace(node.singlechar, "_")

        # TODO: handle node.nocase
        return (
            f"{lhs} {'NOT ' if node.not_ else ''}LIKE "
            f"'{pattern}' ESCAPE '{node.escapechar}'"
        )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return f"{lhs} {'NOT ' if node.not_ else ''}IN ({', '.join(options)})"

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return f"{lhs} IS {'NOT ' if node.not_ else ''}NULL"

    # @handle(ast.TemporalPredicate, subclasses=True)
    # def temporal(self, node, lhs, rhs):
    #     pass

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        func = SPATIAL_COMPARISON_OP_MAP[node.op]
        return f"{func}({lhs},{rhs})"

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        func = SPATIAL_COMPARISON_OP_MAP[ast.SpatialComparisonOp.INTERSECTS]
        rhs = f"ST_GeomFromText('POLYGON(({node.minx} {node.miny}, {node.minx} {node.maxy}, {node.maxx} {node.maxy}, {node.maxx} {node.miny}, {node.minx} {node.miny}))')"  # noqa
        return f"{func}({lhs},{rhs})"

    @handle(ast.Attribute)
    def attribute(self, node: ast.Attribute):
        return f'"{self.attribute_map[node.name]}"'

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node: ast.Arithmetic, lhs, rhs):
        op = ARITHMETIC_OP_MAP[node.op]
        return f"({lhs} {op} {rhs})"

    @handle(ast.Function)
    def function(self, node, *arguments):
        func = self.function_map[node.name]
        return f"{func}({','.join(arguments)})"

    @handle(*values.LITERALS)
    def literal(self, node):
        if isinstance(node, str):
            return f"'{node}'"
        else:
            # TODO:
            return str(node)

    @handle(values.Geometry)
    def geometry(self, node: values.Geometry):
        wkb_hex = shapely.geometry.shape(node).wkb_hex
        return f"ST_GeomFromWKB(x'{wkb_hex}')"

    @handle(values.Envelope)
    def envelope(self, node: values.Envelope):
        wkb_hex = shapely.geometry.box(node.x1, node.y1, node.x2, node.y2).wkb_hex
        return f"ST_GeomFromWKB(x'{wkb_hex}')"


def to_sql_where(
    root: ast.Node,
    field_mapping: Dict[str, str],
    function_map: Optional[Dict[str, str]] = None,
) -> str:
    return SQLEvaluator(field_mapping, function_map or {}).evaluate(root)



import base64
import datetime

from ... import ast
from ...util import parse_datetime, parse_duration
from .base import FESBaseParser
from .util import Element, ParseInput, handle


class FES20Parser(FESBaseParser):
    namespace = "http://www.opengis.net/fes/2.0"

    # @handle('PropertyIsNil')
    # def property_is_nil(self, node: Element, lhs, rhs):
    #     return ast...

    @handle("After")
    def time_after(self, node: Element, lhs, rhs):
        return ast.TimeAfter(lhs, rhs)

    @handle("Before")
    def time_before(self, node: Element, lhs, rhs):
        return ast.TimeBefore(lhs, rhs)

    @handle("Begins")
    def time_begins(self, node: Element, lhs, rhs):
        return ast.TimeBegins(lhs, rhs)

    @handle("BegunBy")
    def time_begun_by(self, node: Element, lhs, rhs):
        return ast.TimeBegunBy(lhs, rhs)

    @handle("TContains")
    def time_contains(self, node: Element, lhs, rhs):
        return ast.TimeContains(lhs, rhs)

    @handle("During")
    def time_during(self, node: Element, lhs, rhs):
        return ast.TimeDuring(lhs, rhs)

    @handle("TEquals")
    def time_equals(self, node: Element, lhs, rhs):
        return ast.TimeEquals(lhs, rhs)

    @handle("TOverlaps")
    def time_overlaps(self, node: Element, lhs, rhs):
        return ast.TimeOverlaps(lhs, rhs)

    @handle("Meets")
    def time_meets(self, node: Element, lhs, rhs):
        return ast.TimeMeets(lhs, rhs)

    @handle("OverlappedBy")
    def time_overlapped_by(self, node: Element, lhs, rhs):
        return ast.TimeOverlappedBy(lhs, rhs)

    @handle("MetBy")
    def time_met_by(self, node: Element, lhs, rhs):
        return ast.TimeMetBy(lhs, rhs)

    @handle("Ends")
    def time_ends(self, node: Element, lhs, rhs):
        return ast.TimeEnds(lhs, rhs)

    @handle("EndedBy")
    def time_ended_by(self, node: Element, lhs, rhs):
        return ast.TimeEndedBy(lhs, rhs)

    @handle("ValueReference")
    def value_reference(self, node: Element):
        return ast.Attribute(node.text)

    @handle("Literal")
    def literal(self, node: Element):
        type_ = node.get("type").rpartition(":")[2]
        value = node.text
        if type_ == "boolean":
            return value.lower() == "true"
        elif type_ in (
            "byte",
            "int",
            "integer",
            "long",
            "negativeInteger",
            "nonNegativeInteger",
            "nonPositiveInteger",
            "positiveInteger",
            "short",
            "unsignedByte",
            "unsignedInt",
            "unsignedLong",
            "unsignedShort",
        ):
            return int(value)
        elif type_ in ("decimal", "double", "float"):
            return float(value)
        elif type_ == "base64Binary":
            return base64.b64decode(value)
        elif type_ == "hexBinary":
            return bytes.fromhex(value)
        elif type_ == "date":
            return datetime.date.fromisoformat(value)
        elif type_ == "dateTime":
            return parse_datetime(value)
        elif type_ == "duration":
            return parse_duration(value)

        # return to string
        return value


def parse(input_: ParseInput) -> ast.Node:
    return FES20Parser().parse(input_)



from datetime import date, datetime, timedelta
from typing import Dict, Union

from lxml import etree

from ... import values
from ...util import parse_datetime, parse_duration
from .util import Element

Temporal = Union[date, datetime, timedelta, values.Interval]


def _parse_time_position(node: Element, nsmap: Dict[str, str]) -> datetime:
    return parse_datetime(node.text)


def _parse_time_instant(node: Element, nsmap: Dict[str, str]) -> datetime:
    position = node.xpath("gml:timePosition", namespaces=nsmap)[0]
    return _parse_time_position(position, nsmap)


def _parse_time_period(node: Element, nsmap: Dict[str, str]) -> values.Interval:
    begin = node.xpath(
        "gml:begin/gml:TimeInstant/gml:timePosition|gml:beginPosition", namespaces=nsmap
    )[0]
    end = node.xpath(
        "gml:end/gml:TimeInstant/gml:timePosition|gml:endPosition", namespaces=nsmap
    )[0]
    return values.Interval(
        _parse_time_position(begin, nsmap),
        _parse_time_position(end, nsmap),
    )


def _parse_valid_time(node: Element, nsmap: Dict[str, str]) -> Temporal:
    return parse_temporal(node[0], nsmap)


def _parse_duration(node: Element, nsmap: Dict[str, str]) -> timedelta:
    return parse_duration(node.text)


PARSER_MAP = {
    "validTime": _parse_valid_time,
    "timePosition": _parse_time_position,
    "TimeInstant": _parse_time_instant,
    "TimePeriod": _parse_time_period,
    "duration": _parse_duration,
}


def is_temporal(node: Element) -> bool:
    return etree.QName(node).localname in PARSER_MAP


def parse_temporal(node: Element, nsmap: Dict[str, str]) -> Temporal:
    parser = PARSER_MAP[etree.QName(node).localname]
    return parser(node, nsmap)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import json
from datetime import datetime
from typing import List, Union, cast

from ... import ast, values
from ...util import parse_datetime, parse_duration
from ...values import Envelope, Geometry

# https://portal.ogc.org/files/96288


COMPARISON_MAP = {
    "eq": ast.Equal,
    "lt": ast.LessThan,
    "lte": ast.LessEqual,
    "gt": ast.GreaterThan,
    "gte": ast.GreaterEqual,
}

SPATIAL_PREDICATES_MAP = {
    "intersects": ast.GeometryIntersects,
    "equals": ast.GeometryEquals,
    "disjoint": ast.GeometryDisjoint,
    "touches": ast.GeometryTouches,
    "within": ast.GeometryWithin,
    "overlaps": ast.GeometryOverlaps,
    "crosses": ast.GeometryCrosses,
    "contains": ast.GeometryContains,
}

TEMPORAL_PREDICATES_MAP = {
    "before": ast.TimeBefore,
    "after": ast.TimeAfter,
    "meets": ast.TimeMeets,
    "metby": ast.TimeMetBy,
    "toverlaps": ast.TimeOverlaps,
    "overlappedby": ast.TimeOverlappedBy,
    "begins": ast.TimeBegins,
    "begunby": ast.TimeBegunBy,
    "during": ast.TimeDuring,
    "tcontains": ast.TimeContains,
    "ends": ast.TimeEnds,
    "endedby": ast.TimeEndedBy,
    "tequals": ast.TimeEquals,
    # 'anyinteract': ast.TimeAnyInteract, # TODO?
}


ARRAY_PREDICATES_MAP = {
    "aequals": ast.ArrayEquals,
    "acontains": ast.ArrayContains,
    "acontainedBy": ast.ArrayContainedBy,
    "aoverlaps": ast.ArrayOverlaps,
}

ARITHMETIC_MAP = {
    "+": ast.Add,
    "-": ast.Sub,
    "*": ast.Mul,
    "/": ast.Div,
}


def walk_cql_json(node: dict, is_temporal: bool = False) -> ast.AstType:  # noqa: C901
    if is_temporal and isinstance(node, str):
        # Open interval
        if node == "..":
            return None

        try:
            return parse_duration(node)
        except ValueError:
            value = parse_datetime(node)

        if value is None:
            raise ValueError(f"Failed to parse temporal value from {node}")

        return value

    if isinstance(node, (str, float, int, bool)):
        return node

    if isinstance(node, list):
        result = [
            cast(datetime, walk_cql_json(sub_node, is_temporal)) for sub_node in node
        ]
        if is_temporal:
            return values.Interval(*result)
        else:
            return result

    assert isinstance(node, dict)

    # check if we are dealing with a geometry
    if "type" in node and "coordinates" in node:
        # TODO: test if node is actually valid
        return Geometry(node)

    elif "bbox" in node:
        return Envelope(*node["bbox"])

    # decode all other nodes
    for name, value in node.items():
        if name in ("and", "or"):
            sub_items = cast(list, walk_cql_json(value))
            return (ast.And if name == "and" else ast.Or).from_items(*sub_items)

        elif name == "not":
            # allow both arrays and objects, the standard is ambigous in
            # that regard
            if isinstance(value, list):
                value = value[0]
            return ast.Not(cast(ast.Node, walk_cql_json(value)))

        elif name in COMPARISON_MAP:
            return COMPARISON_MAP[name](
                cast(ast.ScalarAstType, walk_cql_json(value[0])),
                cast(ast.ScalarAstType, walk_cql_json(value[1])),
            )

        elif name == "between":
            return ast.Between(
                cast(ast.Node, walk_cql_json(value["value"])),
                cast(ast.ScalarAstType, walk_cql_json(value["lower"])),
                cast(ast.ScalarAstType, walk_cql_json(value["upper"])),
                not_=False,
            )

        elif name == "like":
            return ast.Like(
                cast(ast.Node, walk_cql_json(value["like"][0])),
                cast(str, value["like"][1]),
                nocase=value.get("nocase", True),
                wildcard=value.get("wildcard", "%"),
                singlechar=value.get("singleChar", "."),
                escapechar=value.get("escapeChar", "\\"),
                not_=False,
            )

        elif name == "in":
            return ast.In(
                cast(ast.AstType, walk_cql_json(value["value"])),
                cast(List[ast.AstType], walk_cql_json(value["list"])),
                not_=False,
                # TODO nocase
            )

        elif name == "isNull":
            return ast.IsNull(
                walk_cql_json(value),
                not_=False,
            )

        elif name in SPATIAL_PREDICATES_MAP:
            return SPATIAL_PREDICATES_MAP[name](
                cast(ast.SpatialAstType, walk_cql_json(value[0])),
                cast(ast.SpatialAstType, walk_cql_json(value[1])),
            )

        elif name in TEMPORAL_PREDICATES_MAP:
            return TEMPORAL_PREDICATES_MAP[name](
                cast(ast.TemporalAstType, walk_cql_json(value[0], is_temporal=True)),
                cast(ast.TemporalAstType, walk_cql_json(value[1], is_temporal=True)),
            )

        elif name in ARRAY_PREDICATES_MAP:
            return ARRAY_PREDICATES_MAP[name](
                cast(ast.ArrayAstType, walk_cql_json(value[0])),
                cast(ast.ArrayAstType, walk_cql_json(value[1])),
            )

        elif name in ARITHMETIC_MAP:
            return ARITHMETIC_MAP[name](
                cast(ast.ScalarAstType, walk_cql_json(value[0])),
                cast(ast.ScalarAstType, walk_cql_json(value[1])),
            )

        elif name == "property":
            return ast.Attribute(value)

        elif name == "function":
            return ast.Function(
                value["name"],
                cast(List[ast.AstType], walk_cql_json(value["arguments"])),
            )

    raise ValueError(f"Unable to parse expression node {node!r}")


def parse(cql: Union[str, dict]) -> ast.AstType:
    if isinstance(cql, str):
        root = json.loads(cql)
    else:
        root = cql

    return walk_cql_json(root)



from functools import reduce
from operator import add, and_, eq, ge, gt, le, lt, mul, ne, or_, sub, truediv

import shapely

from ...util import like_pattern_to_re


def combine(sub_filters, combinator: str):
    """Combine filters using a logical combinator"""
    assert combinator in ("AND", "OR")
    op = and_ if combinator == "AND" else or_
    return reduce(lambda acc, q: op(acc, q) if acc is not None else q, sub_filters)


def negate(sub_filter):
    """Negate a filter, opposing its meaning."""
    return ~sub_filter


OP_MAP = {
    "<": lt,
    "<=": le,
    ">": gt,
    ">=": ge,
    "<>": ne,
    "=": eq,
}


def compare(lhs, rhs, op):
    return OP_MAP[op](lhs, rhs)


def between(lhs, low, high, not_):
    result = lhs.between(low, high)
    if not_:
        result = ~result
    return result


def like(lhs, pattern, nocase, wildcard, singlechar, escapechar, not_):
    regex = like_pattern_to_re(
        pattern, nocase, wildcard, singlechar, escapechar or "\\"
    )
    result = lhs.str.match(regex)
    if not_:
        result = ~result
    return result


def contains(lhs, items, not_):
    # TODO: check if dataframe or scalar
    result = lhs.isin(items)
    if not_:
        result = ~result
    return result


def null(lhs, not_):
    result = lhs.isnull()
    if not_:
        result = ~result
    return result


def temporal(lhs, time_or_period, op):
    pass
    # TODO implement


SPATIAL_OP_MAP = {
    "INTERSECTS": "intersects",
    "DISJOINT": "disjoint",
    "CONTAINS": "contains",
    "WITHIN": "within",
    "TOUCHES": "touches",
    "CROSSES": "crosses",
    "OVERLAPS": "overlaps",
    "EQUALS": "geom_equals",
}


def spatial(lhs, rhs, op):
    assert op in SPATIAL_OP_MAP
    return getattr(lhs, SPATIAL_OP_MAP[op])(rhs)


def bbox(lhs, minx, miny, maxx, maxy, crs=None):
    box = shapely.geometry.Polygon.from_bounds(minx, miny, maxx, maxy)
    # TODO: handle CRS
    return lhs.intersects(box)


def attribute(df, name, field_mapping=None):
    if field_mapping:
        name = field_mapping[name]
    return df[name]


OP_TO_FUNC = {"+": add, "-": sub, "*": mul, "/": truediv}


def arithmetic(lhs, rhs, op):
    """Create an arithmetic filter"""
    return OP_TO_FUNC[op](lhs, rhs)
